package exercises;

/**
 * Implements an AVL tree. Original author Mark Allen Weiss.
 *
 * Class invariant expressed in Alloy style:
 * all x: AvlNode | x in this.root.*(left + right) - null => 
 *     (x !in x.^(left + right)) && 
 *     (all y: AvlNode | (y in x.left.*(left + right) - null) => y.element < x.element ) && 
 *     (all y: AvlNode | (y in x.right.*(left +right) - null) => x.element < y.element ) && 
 *     (x.left = null && x.right = null => x.height = 0) && 
 *     (x.left = null && x.right != null => x.height = 1 && x.right.height = 0) && 
 *     (x.left != null && x.right = null => x.height = 1 && x.left.height = 0) && 
 *     (x.left != null && x.right != null => x.height = (x.left.height > x.right.height ? x.left.height : x.right.height) + 1 && 
 *                                          (x.left.height > x.right.height ? x.left.height - x.right.height : x.right.height - x.left.height) <=1) 
 */
public class AvlTree
{	/**
	 * Double rotate binary tree node: first left child with its right child; then
	 * node k3 with new left child. For AVL trees, this is a double rotation for
	 * case 2. Update heights, then return new root.
	 */
	private static AvlNode doubleWithLeftChild(final AvlNode k3) {
		k3.left = AvlTree.rotateWithRightChild(k3.left);
		return AvlTree.rotateWithLeftChild(k3);
	}

	/**
	 * Double rotate binary tree node: first right child with its left child; then
	 * node k1 with new right child. For AVL trees, this is a double rotation for
	 * case 3. Update heights, then return new root.
	 */
	private static AvlNode doubleWithRightChild(final AvlNode k1) {
		k1.right = AvlTree.rotateWithLeftChild(k1.right);
		return AvlTree.rotateWithRightChild(k1);
	}

	/**
	 * Return the height of node t, or -1, if null.
	 */
	private static int height(final AvlNode t) {
		return t == null ? -1 : t.height;
	}

	/**
	 * Return maximum of lhs and rhs.
	 */
	private static int max(final int lhs, final int rhs) {
		return lhs > rhs ? lhs : rhs;
	}

	/**
	 * Rotate binary tree node with left child. For AVL trees, this is a single
	 * rotation for case 1. Update heights, then return new root.
	 */
	private static AvlNode rotateWithLeftChild(final AvlNode k2) {
		final AvlNode k1 = k2.left;
		k2.left = k1.right;
		k1.right = k2;
		k2.height = AvlTree.max(AvlTree.height(k2.left), 
		            AvlTree.height(k2.right)) + 1;
		k1.height = AvlTree.max(AvlTree.height(k1.left), k2.height) + 1;
		return k1;
	}

	/**
	 * Rotate binary tree node with right child. For AVL trees, this is a single
	 * rotation for case 4. Update heights, then return new root.
	 */
	private static AvlNode rotateWithRightChild(final AvlNode k1) {
		final AvlNode k2 = k1.right;
		k1.right = k2.left;
		k2.left = k1;
		k1.height = AvlTree.max(AvlTree.height(k1.left), 
		            AvlTree.height(k1.right)) + 1;
		k2.height = AvlTree.max(AvlTree.height(k2.right), k1.height) + 1;
		return k2;
	}

	/** The tree root. */
	private AvlNode root;

	/**
	 * Construct the tree.
	 */
	public AvlTree() {
		this.root = null;
	}

	/**
	 * Find an item in the tree.
	 * 
	 * @param x
	 *          the item to search for.
	 * @return the matching item or null if not found.
	 *
	 * Here is the method specification in Alloy style:
	 * requires: true
	 * modifies: nothing
	 * ensures: (some n: AvlNode | n in this.root.*(left + right) - null && n.element = x)
	 *		? \return in  { n : AvlNode | n in this.root.*(left + right) - null && n.element = x } 
	 *		: \return = null ;
	 */
	public Integer find(final int x) {
		final AvlNode n = find(x, this.root);
		if (n == null) return null;
		else return n.element;
	}

	private AvlNode find(final int x, final AvlNode arg) {
		AvlNode t = arg;
		while (t != null) {
			if (x < t.element) {
				t = t.left;
			} else if (x > t.element) {
				t = t.right;
			} else {
				return t; // Match
			}
		}

		return null; // No match
	}

	/**
	 * Find the largest item in the tree.
	 * 
	 * @return the largest item of null if empty.
	 *
	 * Here is the method specification in Alloy style:
	 * requires: true
	 * modifies: nothing
	 * ensures: some this.root.*(left + right) - null 
	 *		? \return in {n : AvlNode | n in this.root.*(left + right) - null && ( no y in AvlNode | ( y in this.nodes && y.element > n.element ) ) }
	 *		: \return = null ;
	 */
	public Integer findMax() {
		final AvlNode n = findMax(this.root);
		if (n == null) return null;
		else return n.element;
	}

	private AvlNode findMax(final AvlNode arg) {
		AvlNode t = arg;
		if (t == null) {
			return t;
		}

		while (t.right != null) {
			t = t.right;
		}
		return t;
	}

	/**
	 * Find the smallest item in the tree.
	 * 
	 * @return smallest item or null if empty.
	 *
	 * Here is the method specification in Alloy style:
	 * requires: true
	 * modifies: nothing
	 * ensures: some this.root.*(left + right) - null 
	 *		? \return in {n : AvlNode | n in this.root.*(left + right) - null && ( no y in AvlNode | ( y in this.nodes && y.element < n.element ) ) }
	 *		: \return = null ;
	 */
	public Integer findMin() {
		final AvlNode n = findMin(this.root);
		if (n == null) return null;
		else return n.element;
	}

	private AvlNode findMin(final AvlNode arg) {
		AvlNode t = arg;
		if (t == null) {
			return t;
		}

		while (t.left != null) {
			t = t.left;
		}
		return t;
	}

	/**
	 * Insert into the tree; duplicates are ignored.
	 * 
	 * @param x
	 *          the item to insert.
	 *
	 * Here is the method specification in Alloy style:
	 * requires: true
	 * modifies: everything
	 * ensures: (all n : this.root.*(left + right) - null |
	 *		(some n' : \old(this).root.*(left + right) - null | n.element = n'.element)
	 *              or n.element = x)
	 *          and some n : this.root.*(left + right) - null | n.element = x
	 */
	public void insertElem(final int x) {
		this.root = insert(x, this.root);
	}

	private AvlNode insert(final int x, final AvlNode arg) {
		AvlNode t = arg;
		if (t == null) {
			t = new AvlNode(x, null, null);
		} else if (x < t.element) {
			t.left = insert(x, t.left);
			if (AvlTree.height(t.left) - AvlTree.height(t.right) == 2) {
				if (x < t.left.element) {
					t = AvlTree.rotateWithLeftChild(t);
				} else {
					t = AvlTree.doubleWithLeftChild(t);
				}
			}
		} else if (x > t.element) {
			t.right = insert(x, t.right);
			if (AvlTree.height(t.right) - AvlTree.height(t.left) == 2) {
				if (x > t.right.element) {
					t = AvlTree.rotateWithRightChild(t);
				} else {
					t = AvlTree.doubleWithRightChild(t);
				}
			}
		} else {
			; // Duplicate; do nothing
		}
		t.height = AvlTree.max(AvlTree.height(t.left), AvlTree.height(t.right)) + 1;
		return t;
	}

	/**
	 * Test if the tree is logically empty.
	 * 
	 * @return true if empty, false otherwise.
	 *
	 * Here is the method specification in Alloy style:
	 * requires: true
	 * modifies: nothing
	 * ensures: \return = (this.root.*(left + right) = { null })
	 */
	public boolean isEmpty() {
		return this.root == null;
	}

	/**
	 * Make the tree logically empty.
	 *
	 * Here is the method specification in Alloy style:
	 * requires: true
	 * modifies: everything
	 * ensures: this.root.*(left + right) = { null }
	 */
	public void makeEmpty() {
		this.root = null;
	}
}
